@@ -1,6 +1,6 @@
 # License: Simplified BSD, 2014
 # See LICENSE.txt for more information
-from __future__ import print_function, unicode_literals
+from __future__ import absolute_import, print_function, division, unicode_literals
 
 import sys
 import threading
@@ -35,6 +35,9 @@ def _format_operand(o):
     if isinstance(o, string_type):
         return o
 
+    if hasattr(o, '__query_label__'):
+        return o._query_label
+
     if isinstance(o, SciDBArray):
         return o.name
 
