@@ -12,8 +12,8 @@ The following interfaces are currently available:
 """
 # License: Simplified BSD, 2013
 # See LICENSE.txt for more information
+from __future__ import absolute_import, print_function, division, unicode_literals
 
-from __future__ import print_function, unicode_literals
 import warnings
 import abc
 import os
@@ -21,6 +21,8 @@ import atexit
 import logging
 import csv
 from time import time
+from fnmatch import fnmatch
+from zlib import decompress
 
 import requests
 
@@ -32,11 +34,16 @@ import numpy as np
 from .scidbarray import SciDBArray, SciDBDataShape, ArrayAlias, SDB_IND_TYPE
 from .errors import SHIM_ERROR_DICT, SciDBQueryError, SciDBInvalidSession
 from .utils import broadcastable, _is_query, iter_record, _new_attribute_label, as_list
-from .schema_utils import disambiguate
-from .robust import (join, merge, gemm, assert_single_attribute,
-                     boundify, match_chunks,
-                     gesvd, match_chunk_permuted)
-import arithmetic
+from .schema_utils import (disambiguate, as_row_vector, as_column_vector,
+                           zero_indexed, match_dimensions,
+                           assert_single_attribute,
+                           boundify, match_chunks, redimension,
+                           match_chunk_permuted, assert_schema,
+                           right_dimension_pad, left_dimension_pad)
+from .robust import (join, merge, gemm, uniq, gesvd)
+
+from . import arithmetic, relational
+from .parse import _scidb_serialize
 
 __all__ = ['SciDBInterface', 'SciDBShimInterface', 'connect']
 
@@ -44,6 +51,13 @@ SCIDB_RAND_MAX = 2147483647  # 2 ** 31 - 1
 UNESCAPED_QUOTE = re.compile(r"(?<!\\)'")
 
 
+def unzip(payload):
+    try:
+        return decompress(payload, 31)
+    except:
+        raise ValueError("Could not unzip: %s" % repr(payload))
+
+
 def _df(arr, ind):
     if not isinstance(arr, ArrayAlias):
         arr = ArrayAlias(arr)
@@ -56,17 +70,18 @@ def _af(arr, ind):
     return "{{a.a{i}f}}".format(i=ind).format(a=arr)
 
 
-def _to_bytes(arr):
+def _to_bytes(arr, chunk_size=1000):
     """
     Convert a numpy array to a bytestring in SciDB's binary format
     """
-    # very inefficient like this.
+    arr = _scidb_serialize(arr, chunk_size)
 
     # easy case: no strings, SciDB format matches numpy format
     if not any(np.issubdtype(t, np.character) for l, t in arr.dtype.descr):
         return arr.tostring(order='C')
 
     # some attributes are strings
+    # very inefficient like this.
     result = []
     for item in arr.ravel():
         for datum in iter_record(item):
@@ -87,6 +102,7 @@ class SciDBInterface(object):
     def __init__(self):
         self._created = []
         self._persistent = set()
+        self.default_compression = None
         atexit.register(self.reap)
 
     """SciDBInterface Abstract Base Class.
@@ -153,6 +169,19 @@ class SciDBInterface(object):
             self._query_log = []
         self._query_log.append(query)
 
+    @property
+    def default_compression(self):
+        """
+        The default compression to use when downloading data
+        """
+        return self._default_compression
+
+    @default_compression.setter
+    def default_compression(self, value):
+        if value not in [None, 1, 2, 3, 4, 5, 6, 7, 8, 9]:
+            raise ValueError("default_compression must be None or 1-9")
+        self._default_compression = value
+
     @abc.abstractmethod
     def _upload_bytes(self, data):
         """Upload binary data to the SciDB engine
@@ -277,10 +306,8 @@ class SciDBInterface(object):
         datashape = SciDBDataShape.from_schema(schema)
         return SciDBArray(datashape, self, scidbname, persistent=persistent)
 
-    # TODO: give the option to pass a user-defined array name
-    #       (use this in copy(), rename(), and others)
     def new_array(self, shape=None, dtype='double', persistent=False,
-                  **kwargs):
+                  name=None, **kwargs):
         """
         Create a new array, either instantiating it in SciDB or simply
         reserving the name for use in a later query.
@@ -300,6 +327,9 @@ class SciDBInterface(object):
             whether the created array should be persistent, i.e. survive
             in SciDB past when the object wrapper goes out of scope.  Default
             is False.
+        name : str (optional)
+            The name to give the array in the databse. If present,
+            persistent will be set to True.
         **kwargs : (optional)
             If `shape` is specified, additional keyword arguments are passed
             to SciDBDataShape.  Otherwise, these will not be referenced.
@@ -308,8 +338,12 @@ class SciDBInterface(object):
         arr : SciDBArray
             wrapper of the new SciDB array instance.
         """
-        name = self._db_array_name()
-        if shape is not None:
+        if name is not None:
+            persistent = True
+
+        name = name or self._db_array_name()
+
+        if shape is not None or ('dim_low' in kwargs and 'dim_high' in kwargs):
             datashape = SciDBDataShape(shape, dtype, **kwargs)
             query = "CREATE ARRAY {0} {1}".format(name, datashape.schema)
             self._execute_query(query)
@@ -432,6 +466,45 @@ class SciDBInterface(object):
                                                     quotechar=quotechar))
         return arr_list
 
+    def ls(self, pattern='*'):
+        """
+        List the arrays in the database, optionally matching to a pattern
+
+        Parameters
+        ----------
+        pattern : String (optional)
+            A glob-style pattern string. If present, only arrays whose names
+            match the pattern are displayed. '*' matches any string,
+            '?' matches any character
+
+        Returns
+        -------
+        result : list
+            A list of SciDB array names
+        """
+        result = self.list_arrays()
+        return sorted([r for r in result if fnmatch(r, pattern)])
+
+    def remove(self, array):
+        """
+        Remove an array from the database
+
+        This removes the array even if its persistent property is True!
+
+        Parameters
+        ----------
+        array : str or SciDBArray
+            The array (or name of array) to remove
+
+        See Also
+        --------
+        reap(), SciDBArray.reap()
+        """
+        try:
+            self.query("remove({0})", array)
+        except SciDBQueryError:  # array does not exist
+            pass
+
     def ones(self, shape, dtype='double', **kwargs):
         """Return an array of ones
 
@@ -700,34 +773,30 @@ class SciDBInterface(object):
             The wrapper of the SciDB Array, of shape (n, m), consisting of the
             matrix product of A and B
         """
-        # TODO: use GEMM and repartition where applicable.  GEMM requires the
-        #       chunk size to be 32.  We should probably not repartition the
-        #       arrays silently, but instead raise an efficiency warning and
-        #       provide a flag that enables automatic repartitioning.
 
         # TODO: make matrix-vector and vector-vector cases more efficient.
         #       Currently they involve creating copies of the arrays, but this
         #       is just a place-holder for a more efficient implementation.
-        A = boundify(A)
-        B = boundify(B)
-
         if A.ndim not in (1, 2) or B.ndim not in (1, 2):
             raise ValueError("dot requires 1 or 2-dimensional arrays")
 
-        if A.shape[-1] != B.shape[0]:
-            raise ValueError("array dimensions must match for dot product")
-
-        output_shape = A.shape[:-1] + B.shape[1:]
+        A = boundify(A)
+        B = boundify(B)
 
         # TODO: the following four transformations should be done by building
         #       a single query rather than executing separate queries.
         #       The following should be considered a place-holder for right
         #       now.
         if A.ndim == 1:
-            A = A.reshape((1, A.size))
+            A = as_row_vector(A)
 
         if B.ndim == 1:
-            B = B.reshape((B.size, 1))
+            B = as_column_vector(B)
+
+        A = zero_indexed(A)
+        B = zero_indexed(B)
+
+        A, B = match_dimensions(A, B, ((1, 0),))
 
         if A.sdbtype.nullable[0]:
             A = A.substitute(0)
@@ -738,15 +807,20 @@ class SciDBInterface(object):
         # TODO: is there a more efficient way to do this than instantiating
         #       an array of zeros?
         # TODO: use spgemm() when the matrices are sparse
-        C = self.zeros((A.shape[0], B.shape[1]), dtype=A.dtype)
+        kwargs = {'dtype': 'double'}
+        if A.dim_names[0] != B.dim_names[1]:
+            kwargs['dim_names'] = [A.dim_names[0], B.dim_names[1]]
+        C = self.zeros((A.shape[0], B.shape[-1]), **kwargs)
         C = gemm(A, B, C).eval()
 
-        if C.shape == output_shape:
-            return C
-        elif len(output_shape) == 0:
+        if C.size == 1:
             return C[0, 0]
+        if C.shape[0] == 1:
+            return C[0, :]
+        elif C.shape[1] == 1:
+            return C[:, 0]
         else:
-            return C.reshape(output_shape)
+            return C
 
     def svd(self, A, return_U=True, return_S=True, return_VT=True):
         """Compute the Singular Value Decomposition of the array A:
@@ -781,7 +855,7 @@ class SciDBInterface(object):
                 ret.append(gesvd(A, "'%s'" % output))
         return tuple(ret)
 
-    def from_array(self, A, instance_id=0, **kwargs):
+    def from_array(self, A, instance_id=0, chunk_size=1000, **kwargs):
         """Initialize a scidb array from a numpy array
 
         Parameters
@@ -791,6 +865,8 @@ class SciDBInterface(object):
         instance_id : integer
             the instance ID used in loading
             (default=0; see SciDB documentation)
+        chunk_size : integer or list of integers
+            The chunk size of the uploaded SciDBArray. Default=1000
         **kwargs :
             Additional keyword arguments are passed to new_array()
 
@@ -799,22 +875,21 @@ class SciDBInterface(object):
         arr : SciDBArray
             SciDB Array object built from the input array
         """
+        if 'chunk_overlap' in kwargs:
+            raise ValueError("chunk_overlap not supported by from_array. "
+                             "Use schema_utils.rechunk instead")
         q = self.afl.quote
         A = np.asarray(A)
         instance_id = int(instance_id)
-        filename, session_id = self._upload_bytes(_to_bytes(A))
+
+        filename, session_id = self._upload_bytes(_to_bytes(A, chunk_size=chunk_size))
         filename = q(filename)
 
-        # the array gets scrambled if spread across >1 chunk
-        # follow SciDBR and size array to 1 chunk
-        if 'chunk_size' in kwargs:
-            warnings.warn("Ignoring chunk_size. Use redimension instead")
-        kwargs['chunk_size'] = A.shape
-
-        arr = self.new_array(A.shape, A.dtype, **kwargs)
+        arr = self.new_array(A.shape, A.dtype, chunk_size=chunk_size, **kwargs)
         self.afl.load(arr, filename, instance_id,
                       q(arr.sdbtype.bytes_fmt)).eval(store=False)
         self._release_session(session_id)
+
         return arr
 
     def from_dataframe(self, A, instance_id=0, **kwargs):
@@ -845,7 +920,14 @@ class SciDBInterface(object):
         kwargs['dim_names'] = [A_rec.dtype.names[0]]
 
         # redimension the array on the index
-        arr = self.new_array(shape=A_rec.shape,
+        try:
+            dim_low = (m for m in A.index.min())
+            dim_high = (m for m in A.index.max())
+        except TypeError:
+            dim_low = (A.index.min(),)
+            dim_high = (A.index.max(),)
+
+        arr = self.new_array(dim_low=dim_low, dim_high=dim_high,
                              dtype=A_rec.dtype.descr[1:],
                              **kwargs)
         self.afl.redimension_store(A_sdb, arr).eval(store=False)
@@ -922,6 +1004,25 @@ class SciDBInterface(object):
         #       database
         raise NotImplementedError()
 
+    def unique(self, x, is_sorted=False):
+        """
+        Store the unique elements of an array in a new array
+
+        Parameters
+        ----------
+        x : SciDBArray
+            The array to compute unique elements of.
+        is_sorted : bool
+            Whether the array is pre-sorted. If True,
+            x must be a 1D array.
+
+        Returns
+        -------
+        u : SciDBArray
+           The unique elements of x
+        """
+        return uniq(x, is_sorted=is_sorted)
+
     def _apply_func(self, A, func):
         # TODO: new value name could conflict.  How to generate a unique one?
         # TODO: add optional ``out`` argument as in numpy
@@ -1060,9 +1161,7 @@ class SciDBInterface(object):
         """
         return A.substitute(value)
 
-    def merge(self, A, B):
-        """Merge two arrays"""
-        return merge(A, B)
+    merge = staticmethod(relational.merge)
 
     def join(self, *args):
         """
@@ -1081,7 +1180,9 @@ class SciDBInterface(object):
             The remaining arguments are tuples of dimension indices which
             should be joined.
         """
-        B = match_chunk_permuted(B, A, dims).eval()
+        B, A = match_chunk_permuted(B, A, dims)
+        B = B.eval()
+        A = A.eval()
         dims = [_df(arr, index)
                 for dim in dims
                 for arr, index in zip([A, B], dim)]
@@ -1165,7 +1266,7 @@ class SciDBInterface(object):
                     if (s1 == s2):
                         join_indices.append((i1, i2))
                         if (left.chunk_size[i1] != right.chunk_size[i2] or
-                            left.chunk_overlap[i1] != right.chunk_overlap[i2]):
+                                left.chunk_overlap[i1] != right.chunk_overlap[i2]):
                             raise ValueError("join operations require chunk_"
                                              "size/chunk_overlap to match.")
                     elif s1 == 1:
@@ -1281,6 +1382,193 @@ class SciDBInterface(object):
             arr = arr.transpose(permutation)
         return arr
 
+    def concatenate(self, arrays, axis=0):
+        """
+        Concatenate several arrays along a particular dimension.
+
+        This behaves like numpy's concatenate function when the
+        input array dimensions are > the concatenation axis. It
+        behaves differently than numpy when the array dimensions
+        are less than the concatenation axis, in the following way:
+
+        Concatenating 1D arrays along axis=0 behaves like numpy's vstack.
+        Concatenating 1D arrays along axis=1 behaves like numpy's hstack.
+        Concatenating 1D or 2D arrays along axis=2 behaves like dstack.
+
+        Parameters
+        ----------
+        arrays : Sequence of SciDBArrays
+            The arrays to concatenate
+        axis : int, optional (default 0)
+            The dimension to join on. Array shapes must match along
+            all dimensions except this axis.
+
+        Returns
+        -------
+        stacked : SciDBArray
+            A stacked array
+
+        See Also
+        --------
+        hstack(), vstack(), dstack()
+        """
+        assert_schema(arrays, bounded=True, same_attributes=True)
+        slice_at_end = arrays[0].ndim == 1 and axis == 1
+
+        # pad with extra dimensions as needed
+        arrays = [right_dimension_pad(left_dimension_pad(a, 2), axis + 1)
+                  for a in arrays]
+        arrays = list(map(boundify, arrays))
+
+        nd = arrays[0].ndim
+
+        # check for proper shape
+        if any(a.ndim != nd for a in arrays):
+            raise ValueError("All the input arrays must have the same number of dimensions")
+
+        for i, s in enumerate(arrays[0].shape):
+            if i != axis and any(a.shape[i] != s for a in arrays):
+                raise ValueError("all the input array dimensions except "
+                                 "for the concatenation axis must match exactly")
+
+        # compute the proper position along the concatenation dimension
+        # be careful about nonzero origins here
+        joinidx = 0 if (nd == 1 and axis == 1) else axis
+        join = arrays[0].dim_names[joinidx]
+
+        offset = 0
+        result = list(arrays)
+        idx = _new_attribute_label('idx', *arrays)
+
+        for i in range(len(result)):
+            dims = result[i].dim_names + [idx]
+            atts = result[i].att_names
+
+            zero = result[i].datashape.dim_low[joinidx]
+            result[i] = result[i].apply(idx, '{0}-{1}+{2}'.format(join, zero, offset))
+            offset += result[i].datashape.dim_high[joinidx] - zero + 1
+            result[i] = redimension(result[i], dims, atts, dim_boundaries={idx: (0, offset - 1)})
+
+        # merge into one array
+        r = result[0]
+        for x in result[1:]:
+            r = merge(r, x)
+
+        # redimension into proper shape
+        dim_names = [d for d in r.dim_names if d != idx]
+        dim_names[joinidx] = idx
+        r = redimension(r, dim_names, r.att_names)
+
+        # special case for hstack on 1D arrays
+        if slice_at_end:
+            r = r[0]
+
+        return r
+
+    def hstack(self, arrays):
+        """
+        Stack arrays in sequence horizontally (column wise).
+
+        Parameters
+        ----------
+        arrays : Sequence of SciDBArrays
+            The arrays to join. All arrays must have the same
+            shape along all but the second dimension.
+
+        Returns
+        -------
+        stacked : SciDBArray
+            The array formed by stacking the given arrays.
+
+        See Also
+        --------
+        vstack(), dstack(), concatenate()
+        """
+        return self.concatenate(arrays, axis=1)
+
+    def vstack(self, arrays):
+        """
+        Stack arrays in sequence vertically (column wise).
+
+        Parameters
+        ----------
+        arrays : Sequence of SciDBArrays
+            The arrays to join. All arrays must have the same
+            shape along all but the first dimension.
+
+        Returns
+        -------
+        stacked : SciDBArray
+            The array formed by stacking the given arrays.
+
+        See Also
+        --------
+        hstack(), dstack(), concatenate()
+        """
+        return self.concatenate(arrays, axis=0)
+
+    def dstack(self, arrays):
+        """
+        Stack arrays in sequence depth wise (along the third axis).
+
+        Parameters
+        ----------
+        arrays : Sequence of SciDBArrays
+            The arrays to join. All arrays must have the same
+            shape along all but the third dimension.
+
+        Returns
+        -------
+        stacked : SciDBArray
+            The array formed by stacking the given arrays.
+
+        See Also
+        --------
+        hstack(), vstack(), concatenate()
+        """
+        return self.concatenate(arrays, axis=2)
+
+    def percentile(self, a, q, att=None):
+        """
+        Compute the qth percentile of the data along the specified axis
+
+        Parameters
+        ----------
+        a : SciDBArray
+           Input array
+        q : float in the range [0, 100] or a sequence of floats
+           The percentiles to compute
+        att : str, optional
+           The array attribute to compute percentiles for. Defaults to the first
+           attribute
+
+        Returns
+        -------
+        qs : SciDBArray
+           An array with as many elements as q, listing the data value
+           at each percentile
+        """
+        att = att or a.att_names[0]
+
+        q = np.atleast_1d(q)
+        if q.min() < 0 or q.max() > 100:
+            raise ValueError("Percentiles must be in the range [0, 100]")
+
+        f = self.afl
+        sorted = f.sort(a.project(att))
+        sz = sorted.count()[0].astype(np.int)
+
+        # use linear interpolation
+        inds = (q * (sz - 1) / 100.)
+        lo = np.clip(inds.astype(np.int), 0, (sz - 1))
+        hi = np.clip(lo + 1, 0, (sz - 1))
+        lohi = np.hstack((lo, hi))
+        w = 1 - (inds - lo)
+        result = sorted._integer_index(lohi).toarray()
+
+        result = (result[:q.size] * w + result[q.size:] * (1 - w))
+        return result
+
 
 class SciDBShimInterface(SciDBInterface):
 
@@ -1333,8 +1621,11 @@ class SciDBShimInterface(SciDBInterface):
         s.mount('http://', a)
         self._session = s
 
+        self._auth = None
+
         if digest:
-            s.auth = requests.auth.HTTPDigestAuth(user, password)
+            self._auth = requests.auth.HTTPDigestAuth(user, password)
+            s.auth = self._auth
 
         if pam:
             self.login(user, password)
@@ -1368,22 +1659,30 @@ class SciDBShimInterface(SciDBInterface):
                                         "load_library('dense_linear_algebra')",
                                         release=True)
 
-    def _execute_query(self, query, response=False, n=0, fmt='auto'):
+    def _execute_query(self, query, response=False, n=0, fmt='auto', **kwargs):
         # log the query
         SciDBInterface._execute_query(self, query, response, n, fmt)
 
+        # parse compression
+        comp = kwargs.pop('compression', 'auto')
+        if comp == 'auto':
+            comp = self.default_compression
+        if comp is not None:
+            kwargs['compression'] = comp
+        compressed = comp is not None
+
         session_id = self._shim_new_session()
         if response:
             self._shim_execute_query(session_id, query, save=fmt,
-                                     release=False)
+                                     release=False, **kwargs)
 
             if fmt.startswith('(') and fmt.endswith(')'):
                 # binary format
-                result = self._shim_read_bytes(session_id, n)
+                result = self._shim_read_bytes(session_id, n, compressed)
             else:
                 # text format
-                result = self._shim_read_lines(session_id, n)
-            self._shim_release_session(session_id)
+                result = self._shim_read_lines(session_id, n, compressed)
+            self._shim_release_session(session_id, ignore_invalid=True)
         else:
             self._shim_execute_query(session_id, query, release=True)
             result = None
@@ -1398,6 +1697,10 @@ class SciDBShimInterface(SciDBInterface):
 
     def _shim_url(self, keyword, **kwargs):
 
+        if 'compression' in kwargs:
+            # compression implies streaming=2
+            kwargs['stream'] = 2
+
         # add authentication token if needed
         if self._pam_auth is not None:
             kwargs['auth'] = self._pam_auth
@@ -1409,8 +1712,11 @@ class SciDBShimInterface(SciDBInterface):
         return url
 
     def _shim_urlopen(self, url):
+        logging.getLogger(__name__).debug("REQUEST: %s", url)
         try:
-            r = self._session.get(url, verify=False)
+            r = requests.get(url, verify=False, auth=self._auth)
+            # XXX having trouble with hanging streamed requests here
+            #r = self._session.get(url, verify=False)
             r.raise_for_status()
         except requests.HTTPError as e:
             Error = SHIM_ERROR_DICT[r.status_code]
@@ -1439,11 +1745,12 @@ class SciDBShimInterface(SciDBInterface):
         else:
             self._shim_urlopen(url)
 
-    def _shim_execute_query(self, session_id, query, save=None, release=False):
+    def _shim_execute_query(self, session_id, query, save=None, release=False, **kwargs):
         url = self._shim_url('execute_query',
                              id=session_id,
                              query=quote(query.encode('utf-8')),
-                             release=int(bool(release)))
+                             release=int(bool(release)),
+                             **kwargs)
         if save is not None:
             url += "&save={0}".format(quote(save))
 
@@ -1461,7 +1768,7 @@ class SciDBShimInterface(SciDBInterface):
         url = self._shim_url('cancel', id=session_id)
         self._shim_urlopen(url)
 
-    def _shim_read_lines(self, session_id, n):
+    def _shim_read_lines(self, session_id, n, compressed=False):
         url = self._shim_url('read_lines', id=session_id, n=n)
         t0 = time()
         result = self._shim_urlopen(url)
@@ -1471,21 +1778,28 @@ class SciDBShimInterface(SciDBInterface):
         logging.getLogger(__name__).debug("Transfer time: %0.1f sec", dt)
         logging.getLogger(__name__).debug("Payload:       %0.2f MB", pl)
 
+        if compressed:
+            text_result = unzip(text_result)
+
         # the following check is for Py3K compatibility
         if not isinstance(text_result, string_type):
             text_result = text_result.decode('UTF-8')
+
         return text_result
 
-    def _shim_read_bytes(self, session_id, n):
+    def _shim_read_bytes(self, session_id, n, compressed=False):
         url = self._shim_url('read_bytes', id=session_id, n=n)
         t0 = time()
-        result = self._shim_urlopen(url)
-        bytes_result = result.read()
+        bytes_result = self._shim_urlopen(url).read()
+
         dt = time() - t0
         pl = len(bytes_result) / 1048576
         logging.getLogger(__name__).debug("Transfer time: %0.1f sec", dt)
         logging.getLogger(__name__).debug("Payload:       %0.2f MB", pl)
 
+        if compressed:
+            bytes_result = unzip(bytes_result)
+
         return bytes_result
 
     def _shim_upload_file(self, session_id, data):
