@@ -3,7 +3,8 @@
 # License: Simplified BSD, 2014
 # See LICENSE.txt for more information
 
-from __future__ import print_function, division, unicode_literals
+from __future__ import absolute_import, print_function, division, unicode_literals
+
 import warnings
 import re
 from itertools import chain
@@ -16,10 +17,12 @@ from .errors import SciDBError, SciDBForbidden, SciDBQueryError
 
 # Numpy 1.7 meshgrid backport
 from . import parse
-from .utils import meshgrid, slice_syntax, _is_query, _new_attribute_label
-from ._py3k_compat import genfromstr, iteritems, csv_reader, string_type
-from .schema_utils import change_axis_schema
-from .robust import join, cumulate, reshape, thin
+from .utils import (meshgrid, slice_syntax, _is_query,
+                    _new_attribute_label, as_list)
+from ._py3k_compat import genfromstr, iteritems, csv_reader, string_type, dtype as _dtype
+from .schema_utils import change_axis_schema, dimension_rename, new_alias_label
+from . import schema_utils as su
+from .robust import (join, cumulate, reshape, thin, cross_join)
 
 __all__ = ["sdbtype", "SciDBArray", "SciDBDataShape"]
 
@@ -27,7 +30,7 @@ __all__ = ["sdbtype", "SciDBArray", "SciDBDataShape"]
 # Create mappings between scidb and numpy string representations
 # XXX these are partially deprecated, now that nulls/missing values
 # are handled
-_np_typename = lambda s: np.dtype(s).descr[0][1]
+_np_typename = lambda s: _dtype(s).descr[0][1]
 SDB_NP_TYPE_MAP = {'bool': _np_typename('bool'),
                    'float': _np_typename('float32'),
                    'double': _np_typename('float64'),
@@ -84,7 +87,7 @@ def _parse_csv_builtin(txt, dtype):
                  for i, f in enumerate(dtype.descr)]
     if len(new_dtype) == 1:
         new_dtype = new_dtype[0][1]
-    dtype = np.dtype(new_dtype)
+    dtype = _dtype(new_dtype)
 
     return np.array(r, dtype=dtype)
 
@@ -112,7 +115,7 @@ class sdbtype(object):
 
         else:
             try:
-                self.dtype = np.dtype(typecode)
+                self.dtype = _dtype(typecode)
                 self.schema = None
             except:
                 self.schema = self._regularize(typecode)
@@ -163,8 +166,8 @@ class sdbtype(object):
     @classmethod
     def from_full_rep(cls, rep):
         schema = '<%s>' % (','.join('%s: %s %s' %
-                          (nm, typ, 'NULL' if null else '')
-            for nm, typ, null in rep))
+                                    (nm, typ, 'NULL' if null else '')
+                                    for nm, typ, null in rep))
         return cls(schema)
 
     @classmethod
@@ -213,9 +216,9 @@ class sdbtype(object):
         """
         sdbL = cls._schema_to_list(schema)
         if len(sdbL) == 1:
-            return np.dtype(SDB_NP_TYPE_MAP[sdbL[0][1]])
+            return _dtype(SDB_NP_TYPE_MAP[sdbL[0][1]])
         else:
-            return np.dtype([(s[0], SDB_NP_TYPE_MAP[s[1]]) for s in sdbL])
+            return _dtype([(s[0], SDB_NP_TYPE_MAP[s[1]]) for s in sdbL])
 
     @classmethod
     def _dtype_to_schema(cls, dtype):
@@ -231,11 +234,11 @@ class sdbtype(object):
         schema : string
             a SciDB type schema, for example "<val:double,rank:int32>"
         """
-        dtype = np.dtype(dtype).descr
+        dtype = _dtype(dtype).descr
 
         # Hack: if we re-encode this as a dtype, then de-encode again, numpy
         #       will add default names where they are missing
-        dtype = np.dtype(dtype).descr
+        dtype = _dtype(dtype).descr
         pairs = ["{0}:{1}".format(d[0], _sdb_type(d[1])) for d in dtype]
         return '<{0}>'.format(','.join(pairs))
 
@@ -247,7 +250,6 @@ class SciDBDataShape(object):
     def __init__(self, shape, typecode, dim_names=None,
                  chunk_size=1000, chunk_overlap=0, dim_low=None,
                  dim_high=None):
-
         # Process array shape
         if shape is not None:
             try:
@@ -335,8 +337,18 @@ class SciDBDataShape(object):
 
     @property
     def ndim(self):
+        """
+        The number of dimensions in the array
+        """
         return len(self.dim_names)
 
+    @property
+    def natt(self):
+        """
+        The number of attributes in the array
+        """
+        return len(self.sdbtype.names)
+
     @classmethod
     def from_schema(cls, schema):
         """Create a DataShape object from a SciDB Schema.
@@ -436,7 +448,7 @@ class SciDBDataShape(object):
         dct = dict(f[:2] for f in self.sdbtype.full_rep)
         types = [SDB_NP_TYPE_MAP[dct.get(key, SDB_IND_TYPE)]
                  for key in keys]
-        return np.dtype(list(zip(map(str, keys), types)))
+        return _dtype(list(zip(keys, types)))
 
 
 class ArrayAlias(object):
@@ -465,8 +477,12 @@ class ArrayAlias(object):
             # exception text copied from Python2.6
             raise AttributeError("%r object has no attribute %r" %
                                  (type(self).__name__, attr))
-
         groups = match.groups()
+
+        if groups[1] == '':
+            raise AttributeError("%r object has no attribute %r" %
+                                 (type(self).__name__, attr))
+
         i = int(groups[1])
 
         if groups[2]:
@@ -570,18 +586,11 @@ class SciDBArray(object):
         if self.shape is not None:
             n = min(n, self.size)
 
-        args = zip([0] * (self.ndim - 1), [1] * (self.ndim - 1))
-        result = self.subarray(0, n - 1, *args)
-        try:
-            return result.todataframe()
-        except ImportError:
-            return result.toarray()
+        # use first dimension only
+        args = list(self.datashape.dim_low)
+        args += ([args[0] + n - 1] + args[1:])
+        result = self.between(*args)
 
-    def tail(self, n=5):
-        hi = self.size - 1
-        lo = max(hi - n + 1, 0)
-        args = zip([0] * (self.ndim - 1), [1] * (self.ndim - 1))
-        result = self.subarray(lo, hi, *args)
         try:
             return result.todataframe()
         except ImportError:
@@ -708,6 +717,10 @@ class SciDBArray(object):
     def ndim(self):
         return self.datashape.ndim
 
+    @property
+    def natt(self):
+        return self.datashape.natt
+
     @property
     def size(self):
         return np.prod(self.shape)
@@ -905,7 +918,7 @@ class SciDBArray(object):
                 result = np.zeros(self.shape, self.dtype)
                 coords = tuple([arr[d] for d in self.dim_names])
 
-                if len(self.att_names) == 1:
+                if self.natt == 1:
                     result[coords] = arr[self.att_names[0]]
                     return result
 
@@ -941,11 +954,28 @@ class SciDBArray(object):
 
         return arr
 
-    def toarray(self, transfer_bytes=False):
+    def toarray(self, **kwargs):
         """Transfer data from database and store in a numpy array.
 
         Parameters
         ----------
+        compression : None, 'auto' or 1-9
+           Whether to use compression. None disables compression.
+           'auto' uses the `default_compression` attribute on
+           the SciDB interface object. 1-9 uses gzip compression
+           at the specified level (1=fast, 9=best)
+
+        method : 'sparse' or 'dense' (optional, default sparse)
+            Whether the array to download is sparse or dense.
+
+            'sparse' works with all SciDB arrays, but
+            is slower (it computes and transfers array indices
+            for each cell). It is the default.
+
+            'dense' transfer only works for bound arrays
+            with no empty cells. It is faster, since it
+            doesn't compute or transfer indices.
+
         transfer_bytes : DEPRECATED
            Unused
 
@@ -959,12 +989,13 @@ class SciDBArray(object):
         If the array is backed by a query, the query is evaluated and stored
         in the database
         """
-        if transfer_bytes:
+        if 'transfer_bytes' in kwargs:
             warnings.warn(DeprecationWarning("transfer_bytes is deprecated, "
-                          "and will be removed in a future version"))
+                                             "and will be removed in a future version"))
+            kwargs.pop('transfer_bytes')
 
         self.eval()  # evaluate if needed, for speed
-        return parse.toarray(self)
+        return parse.toarray(self, **kwargs)
 
     def eval(self, out=None, store=True, **kwargs):
         """
@@ -1000,17 +1031,15 @@ class SciDBArray(object):
         self.name = name
         return result
 
-    def todataframe(self, transfer_bytes=True):
+    def todataframe(self, **kwargs):
         """Transfer array from database and store in a local Pandas dataframe
 
-        For multidimensional arrays, the dimension values are added
-        as additional columns in the dataframe.
+        The array dimensions are assigned to the index of the output.
 
         Parameters
         ----------
-        transfer_bytes : boolean
-            if True (default), then transfer data as bytes rather than as
-            ASCII.
+        compression : 'auto', None, or [1-9]
+           Whether and how to compress the transfer.
 
         Returns
         -------
@@ -1018,22 +1047,17 @@ class SciDBArray(object):
             The dataframe object containing the data in the array.
         """
         from pandas import DataFrame
-        if self.ndim == 1:
-            return DataFrame(self.toarray())
 
         idx = _new_attribute_label('row', self)
-        return self.afl.unpack(self, idx).todataframe()
+        a = self.afl.unpack(self, idx).toarray(**kwargs)
+        return DataFrame(a, columns=self.dim_names + self.att_names).set_index(self.dim_names)
 
-    def tosparse(self, sparse_fmt='recarray', transfer_bytes=True):
+    def tosparse(self, sparse_fmt='recarray', **kwargs):
         """Transfer array from database and store in a local sparse array.
 
         Parameters
         ----------
-        transfer_bytes : boolean
-            if True (default), then transfer data as bytes rather than as
-            ASCII.  This is more accurate, but requires two passes over
-            the data (one for indices, one for values).
-        sparse_format : string or None
+        sparse_fmt : string or None
             Specify the sparse format to use.  Available formats are:
             - 'recarray' : a record array containing the indices and
               values for each data point.  This is valid for arrays of
@@ -1042,41 +1066,106 @@ class SciDBArray(object):
               These are valid only for 2-dimensional arrays with a single
               attribute.
 
+        compression : 'auto', None, or [1-9]
+            Whether to use compression. None disables compression.
+            'auto' uses the value from the SciDBInterface's default_compression
+            attribute. 1-9 specifies a gzip-compression level (1=fast, 9=best)
+
+        transfer_bytes : deprecated
+            Unused
+
         Returns
         -------
         arr : ndarray or sparse matrix
             The sparse representation of the data
         """
-        if sparse_fmt == 'recarray':
-            spmat = None
-        else:
-            from scipy import sparse
-            try:
-                spmat = getattr(sparse, sparse_fmt + "_matrix")
-            except AttributeError:
-                raise ValueError("Invalid matrix format: "
-                                 "'{0}'".format(sparse_fmt))
+        if 'transfer_bytes' in kwargs:
+            raise DeprecationWarning("transfer_bytes is deprecated")
+            kwargs.pop('transfer_bytes')
 
-        columns = self._download_data(transfer_bytes=transfer_bytes,
-                                      output='sparse')
+        return parse.tosparse(self, sparse_fmt=sparse_fmt, **kwargs)
 
-        if sparse_fmt == 'recarray':
-            return columns
-        else:
-            from scipy import sparse
-            full_dtype = self.datashape.ind_attr_dtype
-            if self.ndim != 2:
-                raise ValueError("Only recarray format is valid for arrays "
-                                 "with ndim != 2.")
-            if len(full_dtype.names) > 3:
-                raise ValueError("Only recarray format is valid for arrays "
-                                 "with multiple attributes.")
+    def tolist(self, **kwargs):
+        """
+        Download the array as a (nested) python list
+        """
+        return self.toarray(**kwargs).tolist()
 
-            labels = full_dtype.names
-            data = columns[labels[2]]
-            ij = (columns[labels[0]], columns[labels[1]])
-            arr = sparse.coo_matrix((data, ij), shape=self.shape)
-            return spmat(arr)
+    def _integer_index(self, idx):
+        """
+        Index using an integer array
+        """
+        from .schema_utils import disambiguate, redimension
+
+        if self.ndim != 1 or idx.ndim != 1:
+            raise NotImplementedError("Slicing with integers requires 1D arrays")
+
+        if not isinstance(idx, SciDBArray):
+            idx = self.interface.from_array(idx)
+
+        if idx.natt != 1:
+            raise ValueError("Can only index with single-attribute arrays")
+
+        # example array: [0, 10, 20]
+        #         idx  : [0, 2, 0]
+
+        # relabel idx to avoid name collisions
+        self, idx = disambiguate(self, idx)
+
+        idx_att = idx.att_names[0]
+        orig_atts = self.att_names
+
+        # turn into 2D array, 2nd dim = idx
+        # idx: [(0, 0), (2, 1), (0, 2)]
+        pos = _new_attribute_label('pos', self, idx)
+        idx = idx.apply(pos, idx.dim_names[0])
+        # idx: [[(0), (), ()],
+        #       [(), (), (1)],
+        #       [(2), (), ()]]
+        idx = redimension(idx, idx.dim_names + idx.att_names[0:1], [pos])
+
+        #    (value, idx)
+        # x: [[(0,0), (),    (0,2)],
+        #     [(),    (),    ()],
+        #     [(),    (20,1),()]]
+        x = cross_join(self, idx, self.dim_names[0], idx_att)
+
+        # for some reason, SciDB needs to eval() the query for
+        # multiattribute arrays
+        if len(orig_atts) > 1:
+            x = x.eval()
+        x = redimension(x, [pos], orig_atts)
+        return x
+
+    def isel(self, **kwargs):
+        """
+        Select a subset of the array by dimension name
+
+        Parameters
+        ----------
+        kwargs : dimension names -> slice descrption
+           What to select from the array
+
+        Returns
+        -------
+        subarray : SciDBArray
+            The array subset
+
+        Examples
+        --------
+        x = sdb.arange(20).reshape((4, 5))
+        print(x.schema)  # <f0:int64> [i0=0:3,1000,0,i1=0:4,1000,0]
+        x.isel(i0=0)  # x[0]
+        x.isel(i1=2)  # x[:, 2]
+        x.isel(i1=slice(2,4)).toarray()  # x[:, 2:4]
+        """
+        slices = [slice(None) for _ in range(self.ndim)]
+        for k, v in kwargs.items():
+            if k not in self.dim_names:
+                raise KeyError("Invalid dimension name: %s" % k)
+            idx = self.dim_names.index(k)
+            slices[idx] = v
+        return self[slices]
 
     def __getitem__(self, indices):
         # The goal of getitem is to make a numpy-style interface perform
@@ -1091,14 +1180,29 @@ class SciDBArray(object):
         #         specified place in each dimension.
         #  [reshape: This applies if a slice argument is newaxis.]
 
-        # TODO: make this more efficient by using a single query
         # TODO: allow newaxis to be passed
 
         # slice can be either a tuple/iterable or a single integer/slice
 
-        if isinstance(indices, SciDBArray):
+        def _boolean_scidbarray(x):
+            if not isinstance(x, SciDBArray):
+                return False
+            rep = x.sdbtype.full_rep
+            return len(rep) == 1 and rep[0][1] == 'bool'
+
+        def _integer_scidbarray(x):
+            if not isinstance(x, SciDBArray):
+                return False
+            rep = x.sdbtype.full_rep
+            return len(rep) == 1 and rep[0][1].startswith('int')
+
+        # boolean scidbarray
+        if _boolean_scidbarray(indices):
             return self._boolean_filter(indices)
 
+        if _integer_scidbarray(indices):
+            return self._integer_index(indices)
+
         # passing a boolean mask
         if isinstance(indices, np.ndarray) and indices.dtype == np.bool:
             indices = self.interface.from_array(indices)
@@ -1147,7 +1251,8 @@ class SciDBArray(object):
             arr1 = self
 
         # pull out the indices from the remaining slices
-        shape = arr1.shape
+        shape = su.coerced_shape(arr1)
+
         indices = [i for i in indices if isinstance(i, slice)]
         indices = [sl.indices(sh) for sl, sh in zip(indices, shape)]
 
@@ -1189,6 +1294,85 @@ class SciDBArray(object):
         args = [s.start for s in slices] + [s.stop - 1 for s in slices]
         return self.afl.subarray(self, *args)
 
+    def dimension_rename(self, *args):
+        """
+        Rename a set of dimensions
+
+        Parameters
+        ----------
+        args : (old_name, new_name, ...)
+            0 or more rename pairs
+
+        Returns
+        -------
+        renamed : SciDBArray
+            The new array
+        """
+        if len(args) == 0:
+            return self
+        return dimension_rename(self, *args)
+
+    def attribute_rename(self, *args):
+        """
+        Rename a set of attributes
+
+        Parameters
+        ----------
+        args : (old_name, new_name, ...)
+            0 or more rename pairs
+
+        Returns
+        -------
+        renamed : SciDBArray
+            The new array
+        """
+
+        if len(args) == 0:
+            return self
+        return self.afl.attribute_rename(self, *args)
+
+    def relabel(self, renames):
+        """
+        relabel the attributes or dimensions in an array.
+
+        Parameters
+        ----------
+        renames: dict
+            A dictionary mapping old names to new names
+
+        Returns
+        -------
+        renamed : SciDBArray
+            A new array
+        """
+        return su._relabel(self, renames)
+
+    def index_lookup(self, idx_array, attribute, output_attribute='idx'):
+        """ Wrapper around the index_lookup AFL call.
+
+        This automatically wraps the array name and attribute in an alias,
+        as is required by AFL
+
+        Parameters
+        ----------
+        idx_array : SciDBArray
+           A single-attribute array of unique values to lookup
+        attribute : string
+           The name of an attribute in this array
+        output_attribute : string
+           The attribute of the output containing the indices
+
+        Returns
+        -------
+        indexed : SciDBArray
+           The current array appended with an index attribute
+        """
+        output_attribute = _new_attribute_label(output_attribute, self)
+        alias = new_alias_label('x', self)
+        attribute = '%s.%s' % (alias, attribute)
+
+        return self.afl.index_lookup(self.as_(alias), idx_array, attribute, output_attribute)
+
     # join operations: note that these ignore all but the first attribute
     # of each array.
     def __add__(self, other):
@@ -1263,7 +1447,7 @@ class SciDBArray(object):
         if isinstance(other, SciDBArray):
             return self._boolean_compare_array(operator, other)
 
-        if len(self.att_names) > 1:
+        if self.natt > 1:
             raise TypeError("Inequality comparison not supported for "
                             "multi-attribute arrays")
 
@@ -1285,7 +1469,7 @@ class SciDBArray(object):
         """
         from .schema_utils import disambiguate
 
-        if len(self.att_names) > 1 or len(other.att_names) > 1:
+        if self.natt > 1 or other.natt > 1:
             raise TypeError("Inequality comparison not supported for "
                             "multi-attribute arrays")
 
@@ -1321,9 +1505,21 @@ class SciDBArray(object):
         self, mask = disambiguate(self, mask)
         joined = join(self, mask)
         expr = '%s=TRUE' % mask.att_names[0]
-        idx = _new_attribute_label('__idx', self, mask)
-        return f.project(f.unpack(f.filter(joined, expr), idx),
-                         *self.att_names)
+        result = f.project(f.filter(joined, expr), *self.att_names)
+        return result
+
+    def collapse(self):
+        """
+        Flatten and remove all the empty cells.
+
+        Returns
+        -------
+        collapsed : SciDBArray
+           A new 1D dense array, containing all of the nonempty
+           cells in this array.
+        """
+        idx = _new_attribute_label('__idx', self)
+        return self.unpack(idx).project(*self.att_names)
 
     def __lt__(self, other):
         return self._boolean_compare('<', other)
@@ -1343,8 +1539,11 @@ class SciDBArray(object):
     def __gt__(self, other):
         return self._boolean_compare('>', other)
 
+    def __neg__(self):
+        return self * -1
+
     def __invert__(self):
-        if len(self.att_names) != 1:
+        if self.natt != 1:
             raise TypeError("Can only invert single-attribute arrays")
         if self.sdbtype.full_rep[0][1] != 'bool':
             raise TypeError("Can only invert boolean arrays")
@@ -1410,6 +1609,21 @@ class SciDBArray(object):
     # This allows the transpose of A to be computed via A.T
     T = property(transpose)
 
+    def unpack(self, name='idx'):
+        """
+        Unpack with automatic dimension name disambiguation
+
+        Unpacking flattens an array to 1D, converting all old dimensions
+        to attributes
+
+        Parameters
+        ----------
+        name : str (optional, default 'idx')
+           The name of the new dimension. Will be disambiguated
+        """
+        name = _new_attribute_label(name, self)
+        return self.afl.unpack(self, name)
+
     def reshape(self, shape, **kwargs):
         """Reshape data into a new array
 
@@ -1467,6 +1681,37 @@ class SciDBArray(object):
         q = self.afl.substitute(self, b)
         return q
 
+    def aggregate(self, *args, **kwargs):
+        """
+        Perform one or more aggregations over an array.
+
+        Parameters
+        ----------
+        *args: One or more SciDB aggregate expressions
+            Aggregations to perform, like 'sum(value) as x'
+        by :
+
+        Examples
+        --------
+        x = sdb.arange(10).reshape((5, 2))
+        x.aggregate('count(*)').toarray()
+        x.aggregate('max(f0)', by='i0').toarray()
+
+        See Also
+        --------
+        groupby()
+        """
+        by = kwargs.pop('by', None)
+        if kwargs:
+            raise ValueError("Invalid keywords: %s" % list(kwargs.keys()))
+
+        if by is None:
+            return self.afl.aggregate(self, *args)
+        else:
+            by = as_list(by)
+            # XXX deal with aggregation on attributes
+            return self.afl.aggregate(self, *(list(args) + by))
+
     def groupby(self, by):
         """
         Build a groupby object from this array
@@ -1943,6 +2188,34 @@ class SciDBArray(object):
         q = f.redimension(q, schema)
         return q
 
+    def any(self):
+        """
+        Returns whether any elements of each attribute are true.
+
+        Returns
+        -------
+        any : SciDBArray
+            boolean array
+        """
+        if not all(dt == 'bool' for nm, dt, null in self.sdbtype.full_rep):
+            raise TypeError("any() only valid for boolean arrays")
+
+        return self.aggregate(*('max(%s)' % att for att in self.att_names)) > 0
+
+    def all(self):
+        """
+        Returns whether all elements of each attribute are true.
+
+        Returns
+        -------
+        all : SciDBArray
+            boolean array
+        """
+        if not all(dt == 'bool' for nm, dt, null in self.sdbtype.full_rep):
+            raise TypeError("any() only valid for boolean arrays")
+
+        return self.aggregate(*('min(%s)' % att for att in self.att_names)) > 0
+
 
 def _subarray(array, *masks):
     """
