@@ -2,33 +2,21 @@
 
 # License: Simplified BSD, 2014
 # See LICENSE.txt for more information
+from __future__ import absolute_import, print_function, division
 from operator import lt, le, eq, gt, ge, ne
 
 import pytest
 import numpy as np
 from numpy.testing import assert_allclose, assert_array_equal
 
-MISSING_PD = False
-MISSING_SP = False
-try:
-    import pandas as pd
-except ImportError:
-    MISSING_PD = True
-try:
-    from scipy import sparse
-except ImportError:
-    MISSING_SP = True
-
-needs_pandas = pytest.mark.skipif(MISSING_PD, reason='Test requires Pandas')
-needs_scipy = pytest.mark.skipif(MISSING_SP, reason='Test requires SciPy')
-
 
 # In order to run tests, we need to connect to a valid SciDB engine
 from scidbpy import SciDBArray, SciDBShimInterface, connect, SciDBDataShape
-from scidbpy.schema_utils import disambiguate
-from scidbpy.robust import rechunk
+from scidbpy.schema_utils import disambiguate, rechunk
+from scidbpy.robust import join
 
-from . import sdb, TestBase, teardown_function
+from . import (sdb, TestBase, teardown_function,
+               randarray, needs_pandas, needs_scipy, pd, sparse)
 RTOL = 1E-6
 
 
@@ -90,11 +78,21 @@ def test_to_sparse():
     assert_allclose(X, Xcsr.toarray())
 
 
+def test_to_sparse_recarray():
+    """Test export to Scipy Sparse matrix"""
+
+    X = np.random.random((10, 6))
+    Xsdb = sdb.from_array(X)
+    Xrec = Xsdb.tosparse('recarray')
+
+    assert_array_equal(Xsdb.unpack().toarray(), Xrec)
+
+
 @needs_pandas
 def test_from_dataframe():
     """Test import from Pandas dataframe"""
 
-    X = np.zeros(10, dtype=[('i', int), ('f', float)])
+    X = np.zeros(10, dtype=[(str('i'), int), (str('f'), float)])
     X['i'] = np.arange(10)
     X['f'] = 0.1 * np.arange(10, 20)
     Xdf = pd.DataFrame(X)
@@ -102,6 +100,22 @@ def test_from_dataframe():
     assert_array_equal(X, Xsdb.toarray())
 
 
+@needs_pandas
+def test_from_dataframe_nonzero_index():
+    x = pd.DataFrame({'x': [1, 2, 3], 'y': [2, 3, 4]}).set_index('x')
+    xsdb = sdb.from_dataframe(x)
+
+    assert xsdb.dim_names == ['x']
+
+
+@needs_pandas
+def test_from_dataframe_negative_index():
+    x = pd.DataFrame({'x': [-1, 2, 3], 'y': [2, 3, 4]}).set_index('x')
+    xsdb = sdb.from_dataframe(x)
+
+    assert xsdb.dim_names == ['x']
+
+
 @needs_pandas
 def test_to_dataframe():
     """Test export to Pandas dataframe"""
@@ -121,8 +135,7 @@ def test_to_dataframe():
 def test_to_dataframe_multidim():
     x = sdb.afl.build('<a:float>[i=0:2,10,0, j=0:1,10,0]', 'i+j')
 
-    df = x.todataframe()
-
+    df = x.todataframe().reset_index()
     assert_array_equal(df['i'], [0, 0, 1, 1, 2, 2])
     assert_array_equal(df['j'], [0, 1, 0, 1, 0, 1])
     assert_array_equal(df['a'], [0, 1, 1, 2, 2, 3])
@@ -313,30 +326,6 @@ def test_svd():
     assert_allclose(VT.toarray(), VT2, rtol=RTOL)
 
 
-def test_slicing():
-    # note that slices must be a divisor of chunk size
-
-    def check_subarray(slc):
-        A = sdb.random((10, 10), chunk_size=12)
-        Aslc = A[slc]
-        if isinstance(Aslc, SciDBArray):
-            Aslc = Aslc.toarray()
-        assert_allclose(Aslc, A.toarray()[slc], rtol=RTOL)
-
-    for slc in [(slice(None), slice(None)),
-                (2, 3),
-                1,
-                slice(2, 6),
-                (slice(None), 2),
-                (slice(2, 8), slice(3, 7)),
-                (slice(2, 8, 2), slice(None, None, 3)),
-                (slice(2, -2), slice(3, -2))]:
-        yield check_subarray, slc
-
-    # non-supported case
-    #(slice(8, 2, -1), slice(7, 3, -1))
-
-
 def test_abs():
     A = sdb.random((5, 5))
     B = abs(A - 1)
@@ -426,30 +415,6 @@ def test_transpose():
     assert_allclose(A.T.toarray(), A.toarray().T)
 
 
-def test_join():
-    A = sdb.randint(10)
-    B = sdb.randint(10)
-    C = sdb.join(A, B)
-
-    Cnp = C.toarray()
-    names = Cnp.dtype.names
-    assert_allclose(Cnp[names[0]], A.toarray())
-    assert_allclose(Cnp[names[1]], B.toarray())
-
-
-def test_cross_join():
-    A = sdb.random((10, 5))
-    B = sdb.random(10)
-    AB = sdb.cross_join(A, B, (0, 0))
-
-    ABnp = AB.toarray()
-    names = ABnp.dtype.names
-    assert_allclose(ABnp[names[0]],
-                    A.toarray())
-    assert_allclose(ABnp[names[1]],
-                    B.toarray()[:, None] + np.zeros(A.shape[1]))
-
-
 def test_regrid():
     A = sdb.random((8, 4))
     Ag = A.regrid(2, "sum")
@@ -688,41 +653,6 @@ def test_toarray_smallchunk():
     assert_allclose(x.toarray(), expected)
 
 
-class TestAttributeAccess(TestBase):
-
-    def test_single(self):
-        x = sdb.arange(5)
-        assert_array_equal(x['f0'].toarray(), [0, 1, 2, 3, 4])
-
-    def test_multi(self):
-        x = sdb.arange(5)
-        x = sdb.afl.apply(x, 'f1', 'f0+1')
-        x = sdb.afl.apply(x, 'f2', 'f0+2')
-
-        result = x[['f0', 'f1']].todataframe()
-        assert_array_equal(result['f0'], [0, 1, 2, 3, 4])
-        assert_array_equal(result['f1'], [1, 2, 3, 4, 5])
-
-    def test_add_column(self):
-
-        x = sdb.arange(5)
-        x['f1'] = 'f0+1'
-        assert_array_equal(x['f1'].toarray(), [1, 2, 3, 4, 5])
-
-    def test_schema_updates_with_add_column(self):
-        x = sdb.arange(5)
-        x['f1'] = 'f0+1'
-        assert x.att_names == ['f0', 'f1']
-
-    def test_add_multi_columns(self):
-        x = sdb.arange(5)
-        x[['f1', 'f2']] = 'f0+1', 'f0+2'
-        assert x.att_names == ['f0', 'f1', 'f2']
-
-        assert_array_equal(x['f1'].toarray(), [1, 2, 3, 4, 5])
-        assert_array_equal(x['f2'].toarray(), [2, 3, 4, 5, 6])
-
-
 class TestInequality(TestBase):
 
     def test_scalar(self):
@@ -800,59 +730,165 @@ class TestInequality(TestBase):
         assert exc.value.args[0] == 'Can only invert single-attribute arrays'
 
 
-class TestBooleanIndexing(TestBase):
-
-    def test_inequality_filter(self):
-
-        def check(y):
-            x = sdb.from_array(y)
-            assert_array_equal((x[x < 5]).toarray(), y[y < 5])
-
-        yield check, np.arange(12)
-        yield check, np.arange(12).reshape(3, 4)
-        yield check, np.arange(12).astype(np.float)
-
-    def test_size_mismatch(self):
-
-        with pytest.raises(ValueError) as exc:
-            x = sdb.arange(5)
-            y = sdb.from_array(np.array([True, False]))
-            x[y]
-
-        assert exc.value.args[0] == 'Shape of mask does not match array: (2,) vs (5,)'
-
-    def test_numpy_boolean_mask(self):
-
-        x = sdb.arange(10)
-        y = np.arange(10) > 8
-
-        assert_array_equal(x[y].toarray(), [9])
-
-    def test_multiattribute(self):
-
-        x = sdb.arange(5)
-        x = sdb.afl.apply(x, 'y', 'f0+1')
-
-        xnp = x.toarray()
-        mnp = np.arange(5) > 3
-
-        m = sdb.from_array(mnp)
-
-        assert_array_equal(xnp[mnp], x[m].toarray())
-
-    def test_name_collisions(self):
-        x = sdb.arange(5)
-        x = sdb.afl.apply(x, 'condition', 'f0+1', '__idx', 'f0+2')
-
-        xnp = x.toarray()
-        mnp = np.arange(5) > 3
-
-        m = sdb.from_array(mnp)
-
-        assert_array_equal(xnp[mnp], x[m].toarray())
-
-
 def test_rechunk_noop():
     x = sdb.zeros(5)
     y = rechunk(x, chunk_size=x.datashape.chunk_size)
     assert y is x
+
+
+def test_unique():
+    def check(x, is_sorted):
+        y = sdb.from_array(x)
+        assert_array_equal(np.unique(x),
+                           sdb.unique(y, is_sorted).toarray())
+
+    yield check, np.array([1, 1, 2, 3]), True
+    yield check, np.array([3, 2, 1, 3]), False
+    yield check, np.random.randint(0, 5, (3, 4)), False
+
+
+@needs_pandas
+def test_dot_index_align():
+
+    x = pd.DataFrame({'x': np.ones(5)},
+                     index=np.arange(5))
+    y = pd.DataFrame({'x': np.ones(5)},
+                     index=np.arange(5) + 1)
+
+    xs = sdb.from_dataframe(x)
+    ys = sdb.from_dataframe(y)
+    actual = sdb.dot(xs, ys)
+
+    x = x['x'].values
+    y = y['x'].values
+    expected = np.dot(x[1:], y[:-1].T)
+
+    assert_allclose(expected, actual, rtol=RTOL)
+
+
+def test_hstack():
+
+    def check(arrays):
+        expected = np.hstack(arrays)
+        actual = sdb.hstack([sdb.from_array(a) for a in arrays])
+        assert_array_equal(expected, actual.toarray())
+
+    rand = np.random.random
+    yield check, [rand(3), rand(4)]
+    yield check, [rand(3), rand(4), rand(5)]
+    yield check, [rand((3, 4)), rand((3, 1))]
+    yield check, [rand((2, 2, 2)), rand((2, 1, 2))]
+    yield check, [rand((2, 2, 2, 2)), rand((2, 1, 2, 2))]
+    yield check, [np.arange(3), np.arange(4)]
+    yield check, [randarray((3, 1), ('<i1', '<i4')),
+                  randarray((3, 1), ('<i1', '<i4'))]
+
+
+def test_vstack():
+
+    def check(arrays):
+        expected = np.vstack(arrays)
+        actual = sdb.vstack([sdb.from_array(a) for a in arrays])
+        assert_array_equal(expected, actual.toarray())
+
+    rand = np.random.random
+    yield check, [rand(3), rand(3)]
+    yield check, [rand((1, 2)), rand((3, 2))]
+    yield check, [rand((2, 2, 2)), rand((1, 2, 2))]
+    yield check, [rand((1, 3, 2, 2)), rand((2, 3, 2, 2))]
+    yield check, [np.arange(3), np.arange(3)]
+    yield check, [randarray((1, 3), ('<i1', '<i4')),
+                  randarray((1, 3), ('<i1', '<i4'))]
+
+
+def test_dstack():
+
+    def check(arrays):
+        expected = np.dstack(arrays)
+        actual = sdb.dstack([sdb.from_array(a) for a in arrays])
+        assert_array_equal(expected, actual.toarray())
+
+    rand = np.random.random
+    # yield check, [rand(3), rand(3)]
+    yield check, [rand((1, 2)), rand((1, 2))]
+    yield check, [rand((2, 1, 1)), rand((2, 1, 2))]
+    yield check, [rand((1, 3, 2, 2)), rand((1, 3, 1, 2))]
+    yield check, [np.arange(3), np.arange(3)]
+    yield check, [randarray((1, 3), ('<i1', '<i4')),
+                  randarray((1, 3), ('<i1', '<i4'))]
+
+
+def test_ls():
+    x = sdb.zeros(5)
+    sdb.ls(x.name) == [x.name]
+    sdb.ls(x.name + '*') == [x.name]
+    sdb.ls('DOES NOT EXIST') == []
+    assert sorted(sdb.ls('*')) == sorted(sdb.list_arrays().keys())
+
+
+def test_remove_array():
+    x = sdb.zeros(5)
+    sdb.remove(x)
+    assert x.name not in sdb.list_arrays()
+
+
+def test_remove_name():
+    x = sdb.zeros(5)
+    sdb.remove(x.name)
+    assert x.name not in sdb.list_arrays()
+
+
+def test_percentile():
+
+    def check(array, percentiles):
+        expected = np.percentile(array, percentiles)
+        actual = sdb.percentile(sdb.from_array(array), percentiles)
+        assert_allclose(expected, actual)
+
+    x = np.arange(11)
+    yield check, x, [50]
+    yield check, x, 50
+    yield check, x, 43.2
+    yield check, x.astype(float), 43.2
+    yield check, x, [1, 5, 100]
+
+
+def test_percentile_attribute_check():
+    x = sdb.arange(5)
+    y = sdb.arange(5) * 2
+    z = join(x, y)
+
+    expected = np.percentile(y.toarray(), 50)
+    actual = sdb.percentile(z, 50, att=z.att_names[1])
+
+    assert_allclose(expected, actual)
+
+
+def test_index_lookup():
+
+    x = sdb.from_array(np.array([5, 5, 5, 3, 5, 3, 2]))
+    idx = sdb.from_array(np.array([2, 3, 5]))
+
+    x = x.index_lookup(idx, x.att_names[0], 'tmp')
+
+    assert_array_equal(x['tmp'].toarray(), [2, 2, 2, 1, 2, 1, 0])
+
+
+class TestHead(TestBase):
+
+    @needs_pandas
+    def test_1d(self):
+        x = sdb.arange(10)
+        assert_array_equal(x.head()['f0'], [0, 1, 2, 3, 4])
+
+    @needs_pandas
+    def test_offset_dim(self):
+        x = pd.DataFrame({'x': [1, 2, 3], 'y': [2, 3, 4]}).set_index('x')
+        x = sdb.from_dataframe(x)
+        assert_array_equal(x.head()['y'], [2, 3, 4])
+
+    @needs_pandas
+    def test_nd(self):
+        x = sdb.arange(12).reshape((4, 3))
+        y = x.head(n=3)
+        assert len(y) == 3
