@@ -1,5 +1,8 @@
 # -*- coding: utf-8 -*-
-from __future__ import print_function, unicode_literals
+# License: Simplified BSD, 2014
+# See LICENSE.txt for more information
+from __future__ import absolute_import, print_function, division, unicode_literals
+
 
 import numpy as np
 from numpy.testing import assert_allclose, assert_array_equal
@@ -32,7 +35,6 @@ class TestScalar(TestBase):
     def check_scalar(self, typ, value, null, expected=None):
         if expected is None:
             expected = [value]
-        print(value, expected, self.scalar(typ, value, null).toarray())
         self.check(self.scalar(typ, value, null), expected)
 
     def check_null(self, typ):
@@ -86,6 +88,11 @@ class TestScalar(TestBase):
             for null in [False, True]:
                 yield self.check_scalar, 'string', "'%s'" % val, null, [val]
 
+    def test_parse_std(self):
+        # regression test
+        x = sdb.ones(5).std().toarray()
+        assert x == 0
+
 
 class TestMultiAttributeArrays(object):
 
@@ -156,3 +163,45 @@ def test_sparse():
     x = sdb.afl.build('<a:int8>[i=0:1,10,0]', 10)
     x = x.redimension('<a:int8>[i=0:2,10,0]')
     assert_array_equal(toarray(x), [10, 10, 0])
+
+
+def test_compression():
+
+    x = sdb.zeros(5, dtype=int)
+    y = toarray(x, compression=9)
+    assert_array_equal(y, [0, 0, 0, 0, 0])
+
+
+def test_auto_compression_options():
+    sdb.default_compression = None
+
+    x = sdb.zeros(5, dtype=int)
+
+    exp = [0, 0, 0, 0, 0]
+    assert_array_equal(toarray(x, compression=None), exp)
+    assert_array_equal(toarray(x, compression='auto'), exp)
+    sdb.default_compression = 1
+    assert_array_equal(toarray(x, compression='auto'), exp)
+    assert_array_equal(toarray(x), exp)
+    assert_array_equal(toarray(x, compression=9), exp)
+    sdb.default_compression = None
+
+
+def test_dense():
+    x = sdb.afl.build('<a:int8>[i=0:100,7,3, j=0:100,10,2]', 'i+j')
+
+    assert_array_equal(x.toarray(method='sparse'), x.toarray(method='dense'))
+
+
+def test_fromarray_chunksize():
+    from . import unfuzzed
+    from_array = unfuzzed['from_array']
+
+    def check(chunk):
+        x = np.arange(24).reshape(2, 3, 4)
+        y = from_array(x, chunk_size=chunk)
+        assert_array_equal(x, y.toarray())
+        assert chunk == tuple(y.datashape.chunk_size)
+
+    for chunk in [(2, 3, 4), (1, 2, 2), (1, 1, 1), (5, 5, 5), (5, 1, 5)]:
+        yield check, chunk
